// Package key implements epoint key pair generation and handling.
//
// An epoint key is an OpenPGP signing key that contains a self-signed
// user id packet which matches
//     "Issuer (<denomination>)"
// or
//     "Holder of <issuer fpr> (<denomination>)"
//
// The OpenPGP DSA key material is generated from a random seed using
// a deterministic algorithm. (The self-signature is not deterministic
// but the key material and thus the fingerprint is.)
// This makes it possible to represent an obligation issuer or holder key
// pair with a few bits of secret random seed.
// (The user id only needs to be set up correctly when the key is uploaded
// to the epoint server, it is not required for signing draft documents.)
package key

import (
	"crypto"
	"crypto/dsa"
	"crypto/openpgp"
	"crypto/openpgp/packet"
	"crypto/rand"
	"crypto/sha1"
	"fmt"
	"io"
	"math/big"
)

// TODO: keep denomination only in issuer key?
// TODO: cleanup

const P = "A4D2B9575C25F0E622B8694387128A793E1AD27D12FFF4B5BA11A37CEFD31C935BCBB0A944581A6E6DA12986FCBA9D666607D71D365C286B9BCB57F6D938BE74982B7D770CE438F03B0A20ABA02E5691458C39D96E6E86AE564176ED1A6DFBAFB6EE7674CC5EDCF9FEB6158471FB3FAB53BA1CE1BA64C5626B9E8585FCEF5D31"
const Q = "FFFFFFFFFFFFFFFFFFFF254EAF9E7916D607AAAF"
const G = "7EA5C898777BE4BB29DCDC47289E718F7274C9CD7E570D3D552F3B3EE43C3DEF7BA68E57786926520CCAC71DBA13F37C4064395D5AF3334A04ABD8CED5E7FF476C661953936E8ADDE96A39D8C4AC1080A2BE3FE863A24B08BD43827E54AFADA72433704EA3C12E50E5BD08C130C68A1402FC20DA79CFE0DE931C414348D32B10"

func PrivKey(r []byte) *dsa.PrivateKey {
	priv := new(dsa.PrivateKey)
	priv.Parameters.P, _ = new(big.Int).SetString(P, 16)
	priv.Parameters.Q, _ = new(big.Int).SetString(Q, 16)
	priv.Parameters.G, _ = new(big.Int).SetString(G, 16)

	x := new(big.Int)
	for {
		h := sha1.New()
		h.Write(r)
		r = h.Sum()
		x.SetBytes(r)
		if x.Sign() == 1 && x.Cmp(priv.Q) < 0 {
			break
		}
		// rarely reachable
	}
	priv.X = x
	priv.Y = new(big.Int)
	priv.Y.Exp(priv.G, x, priv.P)
	return priv
}

func GenKey() (priv *dsa.PrivateKey, err error) {
	x := make([]byte, len(Q)/2)
	_, err = io.ReadFull(rand.Reader, x)
	priv = PrivKey(x)
	return
}

// NewEntity returns an Entity that contains a fresh DSA private key with a
// single identity composed of the given full name, comment and email, any of
// which may be empty but must not contain any of "()<>\x00".
func NewEntity(priv *dsa.PrivateKey, currentTimeSecs int64, name, comment, email string) (e *openpgp.Entity, err error) {
	uid := packet.NewUserId(name, comment, email)
	if uid == nil {
		return nil, fmt.Errorf("NewEntity: invalid argument: user id field contained invalid characters")
	}
	t := uint32(currentTimeSecs)
	e = &openpgp.Entity{
		PrimaryKey: packet.NewDSAPublicKey(t, &priv.PublicKey, false /* not a subkey */ ),
		PrivateKey: packet.NewDSAPrivateKey(t, priv, false /* not a subkey */ ),
		Identities: make(map[string]*openpgp.Identity),
	}
	isPrimaryId := true
	e.Identities[uid.Id] = &openpgp.Identity{
		Name:   uid.Name,
		UserId: uid,
		SelfSignature: &packet.Signature{
			CreationTime: t,
			SigType:      packet.SigTypePositiveCert,
			PubKeyAlgo:   packet.PubKeyAlgoDSA,
			Hash:         crypto.SHA256,
			IsPrimaryId:  &isPrimaryId,
			FlagsValid:   true,
			FlagSign:     true,
			FlagCertify:  true,
			IssuerKeyId:  &e.PrimaryKey.KeyId,
		},
	}
	/*
		e.Subkeys = make([]Subkey, 1)
		e.Subkeys[0] = Subkey{
			PublicKey:  packet.NewRSAPublicKey(t, &encryptingPriv.PublicKey, true),
			PrivateKey: packet.NewRSAPrivateKey(t, encryptingPriv, true),
			Sig: &packet.Signature{
				CreationTime:              t,
				SigType:                   packet.SigTypeSubkeyBinding,
				PubKeyAlgo:                packet.PubKeyAlgoRSA,
				Hash:                      crypto.SHA256,
				FlagsValid:                true,
				FlagEncryptStorage:        true,
				FlagEncryptCommunications: true,
				IssuerKeyId:               &e.PrimaryKey.KeyId,
			},
		}
	*/
	return
}

// simple key generation for obligation issuer clients
func NewIssuerEntity(r []byte, denomination string) (e *openpgp.Entity, err error) {
	return NewEntity(PrivKey(r), 0, "Issuer", denomination, "")
}
// simple key generation for obligation holder clients
func NewHolderEntity(r []byte, issuer, denomination string) (e *openpgp.Entity, err error) {
	return NewEntity(PrivKey(r), 0, "Holder of "+issuer, denomination, "")
}

// check the issuer and denomination associated with the given pgp key
func CheckEntity(e *openpgp.Entity) (isIssuer bool, issuer, denomination string, err error) {
	// TODO: allow non-epoint uids
	if len(e.Identities) != 1 {
		err = fmt.Errorf("CheckEntity: expected one identity")
		return
	}
	for _, i := range e.Identities {
		denomination = i.UserId.Comment
		if i.UserId.Name == "Issuer" {
			isIssuer = true
			issuer = fmt.Sprintf("%X", e.PrimaryKey.Fingerprint)
			return
		}
		prefix := "Holder of "
		if i.UserId.Name[:len(prefix)] == prefix {
			issuer = i.UserId.Name[len(prefix):]
			return
		}
		break
	}
	err = fmt.Errorf("CheckENtity: invalid userid")
	return
}
